import numpy as np
import pytest
from numpy.testing import assert_array_equal, assert_allclose

import menpo.io as mio
from menpo.image.base import (Image, _convert_patches_list_to_single_array,
                              _create_patches_image)
from menpo.image.patches import (extract_patches_with_slice,
                                 extract_patches_by_sampling)
from menpo.shape import PointCloud


#######################
# EXTRACT PATCHES TESTS
#######################
def test_double_type():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert (patches[0].pixels.dtype == np.float64)


def test_float_type():
    image = mio.import_builtin_asset('breakingbad.jpg')
    image.pixels = image.pixels.astype(np.float32)
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert (patches[0].pixels.dtype == np.float32)


def test_uint8_type():
    image = mio.import_builtin_asset('breakingbad.jpg', normalize=False)
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert (patches[0].pixels.dtype == np.uint8)


def test_uint16_type():
    image = Image.init_blank([100, 100], dtype=np.uint16)
    patch_shape = (16, 16)
    landmarks = PointCloud(np.array([[50, 50.]]))
    patches = image.extract_patches(landmarks,
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert (patches[0].pixels.dtype == np.uint16)


def test_int_pointcloud():
    image = Image.init_blank([100, 100])
    patch_shape = (16, 16)
    landmarks = PointCloud(np.array([[50, 50]]))
    patches = image.extract_patches(landmarks,
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert (patches[0].pixels.dtype == np.float)


def test_uint8_type_single_array():
    image = mio.import_builtin_asset('breakingbad.jpg', normalize=False)
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=True)
    assert (patches.dtype == np.uint8)


def test_squared_even_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_squared_odd_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (15, 15)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_nonsquared_even_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 18)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_nonsquared_odd_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (15, 17)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_nonsquared_even_odd_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (15, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_squared_even_patches_landmarks():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches_around_landmarks('PTS',
                                                     patch_shape=patch_shape,
                                                     as_single_array=False)
    assert len(patches) == 68


def test_squared_even_patches_single_array():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    as_single_array=True,
                                    patch_shape=patch_shape)
    assert patches.shape, ((68, 1 == 3) + patch_shape)


def test_squared_even_patches_sample_offsets():
    image = mio.import_builtin_asset('breakingbad.jpg')
    sample_offsets = np.array([[0, 0], [1, 0]])
    patches = image.extract_patches(image.landmarks['PTS'],
                                    sample_offsets=sample_offsets,
                                    as_single_array=False)
    assert len(patches) == 136


@pytest.mark.parametrize('patch_shape,expected_valid', [((15, 13), (8, 7)),
                                                        ((16, 12), (8, 6))],
                         ids=str)
def test_slicing_out_of_bounds(patch_shape, expected_valid):
    image = mio.import_builtin_asset('breakingbad.jpg')
    sample_offsets = np.array([[0, 0], [1, 0]])
    cval = -100
    points = np.array([[0, 0.]])
    sliced_patches = extract_patches_with_slice(
        image.pixels, points, patch_shape, sample_offsets, cval=cval)

    offset = (patch_shape[0] // 2, patch_shape[1] // 2)
    assert_allclose(sliced_patches[0, 0, :, offset[0]:, offset[1]:],
                    image.pixels[:, :expected_valid[0], :expected_valid[1]],
                    rtol=1e-4)
    assert_allclose(sliced_patches[0, 0, :, :offset[0], :offset[1]], -100)


@pytest.mark.parametrize('patch_shape,expected_valid', [((15, 13), (8, 7)),
                                                        ((16, 12), (8, 6))],
                         ids=str)
def test_sampling_out_of_bounds(patch_shape, expected_valid):
    image = mio.import_builtin_asset('breakingbad.jpg')
    sample_offsets = np.array([[0, 0], [1, 0]])
    cval = -100
    points = np.array([[0, 0.]])
    sliced_patches = extract_patches_by_sampling(
        image.pixels, points, patch_shape, sample_offsets, cval=cval)

    offset = (patch_shape[0] // 2, patch_shape[1] // 2)
    assert_allclose(sliced_patches[0, 0, :, offset[0]:, offset[1]:],
                    image.pixels[:, :expected_valid[0], :expected_valid[1]],
                    rtol=1e-4)
    assert_allclose(sliced_patches[0, 0, :, :offset[0], :offset[1]], -100)

    # Offset in row direction by 1
    assert_allclose(sliced_patches[0, 1, :, offset[0]:, offset[1]:],
                    image.pixels[:, 1:expected_valid[0] + 1, :expected_valid[1]],
                    rtol=1e-4)
    assert_allclose(sliced_patches[0, 1, :, :offset[0], :offset[1]], -100)


@pytest.mark.parametrize('patch_shape', [(15, 13), (16, 12)], ids=str)
def test_slicing_equals_sampling(patch_shape):
    image = mio.import_builtin_asset('breakingbad.jpg')
    sample_offsets = np.array([[0, 0], [1, 0]])
    points = image.landmarks['PTS'].points
    cval = -100
    # Add an extra point that is partially out of bounds
    points = np.concatenate([points, [[0, 0]]])
    sliced_patches = extract_patches_with_slice(
        image.pixels, points, patch_shape, sample_offsets, cval=cval)
    sampled_patches = extract_patches_by_sampling(
        image.pixels, points, patch_shape, sample_offsets, order=0, cval=cval)

    assert_allclose(sliced_patches, sampled_patches, rtol=1e-4)
    assert_allclose(sliced_patches[-1, 0, 0, 0, 0], -100)


#######################
# SET PATCHES TESTS
#######################
def test_single_ndarray_patch():
    patch_shape = (8, 7)
    n_channels = 4
    im = Image.init_blank((32, 32), n_channels)
    patch = np.zeros((2, 2, n_channels) + patch_shape)
    patch[0, 0, ...] = np.full((n_channels,) + patch_shape, 1)  # Should be unused
    patch[0, 1, ...] = np.full((n_channels,) + patch_shape, 2)
    patch[1, 0, ...] = np.full((n_channels,) + patch_shape, 3)  # Should be unused
    patch[1, 1, ...] = np.full((n_channels,) + patch_shape, 4)
    patch_center = PointCloud(np.array([[4., 4.], [16., 16.]]))
    new_im = im.set_patches(patch, patch_center, offset_index=1)
    res = np.zeros((32, 32))
    res[:8, 1:8] = 2
    res[12:20, 13:20] = 4
    assert_array_equal(new_im.pixels[2], res)


def test_single_list_patch():
    patch_shape = (8, 7)
    n_channels = 4
    im = Image.init_blank((32, 32), n_channels)
    patch = [Image(np.full((n_channels,) + patch_shape, 1)),
             Image(np.full((n_channels,) + patch_shape, 2)),  # Should be unused
             Image(np.full((n_channels,) + patch_shape, 3)),
             Image(np.full((n_channels,) + patch_shape, 4))]  # Should be unused
    patch_center = PointCloud(np.array([[4., 4.], [16., 16.]]))
    new_im = im.set_patches(patch, patch_center, offset_index=0)
    res = np.zeros((32, 32))
    res[:8, 1:8] = 1
    res[12:20, 13:20] = 3
    assert_array_equal(new_im.pixels[0], res)


def test_offset_argument():
    patch_shape = (5, 6)
    offsets = [(0., 0.), [0., 0.], np.array([[1., 1.]]), None]
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_center = PointCloud(np.array([[100., 101.], [50., 41.]]))
    patch = np.zeros((2, 1, image.n_channels) + patch_shape)
    patch[0, 0, ...] = np.ones((image.n_channels,) + patch_shape)
    patch[1, 0, ...] = 2 * np.ones((image.n_channels,) + patch_shape)
    for off in offsets:
        image = image.set_patches(patch, patch_center, offset=off)
        assert_array_equal(image.pixels[:, 98:103, 98:104], patch[0, 0, ...])
        assert_array_equal(image.pixels[:, 48:53, 38:44], patch[1, 0, ...])


def test_convert_patches_list_to_single_array():
    patch_shape = (7, 2)
    n_channels = 10
    n_centers = 2
    n_offsets = 2
    patches_list = [Image(1 * np.ones((n_channels,) + patch_shape)),
                    Image(2 * np.ones((n_channels,) + patch_shape)),
                    Image(3 * np.ones((n_channels,) + patch_shape)),
                    Image(4 * np.ones((n_channels,) + patch_shape))]
    patches_array = np.zeros((n_centers, n_offsets, n_channels) + patch_shape)
    patches_array[0, 0, ...] = patches_list[0].pixels
    patches_array[0, 1, ...] = patches_list[1].pixels
    patches_array[1, 0, ...] = patches_list[2].pixels
    patches_array[1, 1, ...] = patches_list[3].pixels
    assert_array_equal(
        _convert_patches_list_to_single_array(patches_list, n_centers),
        patches_array)


def test_set_patches_around_landmarks():
    patch_shape = (21, 12)
    image = mio.import_builtin_asset.takeo_ppm()
    patches1 = image.extract_patches_around_landmarks(
        patch_shape=patch_shape, as_single_array=True)
    new_image1 = Image.init_blank(image.shape, image.n_channels)
    new_image1.landmarks['PTS'] = image.landmarks['PTS']
    extracted1 = new_image1.set_patches_around_landmarks(patches1)

    patches2 = image.extract_patches_around_landmarks(
        patch_shape=patch_shape, as_single_array=False)
    new_image2 = Image.init_blank(image.shape, image.n_channels)
    new_image2.landmarks['PTS'] = image.landmarks['PTS']
    extracted2 = new_image2.set_patches_around_landmarks(patches2)
    assert_array_equal(extracted1.pixels, extracted2.pixels)


def test_create_patches_image():
    patch_shape = (7, 14)
    image = mio.import_builtin_asset.takeo_ppm()
    patches = image.extract_patches_around_landmarks(
        patch_shape=patch_shape, as_single_array=True)
    pc = image.landmarks['PTS']
    patches_image = _create_patches_image(patches, pc,
                                          patches_indices=list(range(17)))
    assert (patches_image.n_channels == patches.shape[2])
    assert (patches_image.landmarks.n_groups == 1)
    assert (patches_image.landmarks['patch_centers'].n_points == 17)



from .base import Transform, TransformChain
from .homogeneous import *
from .thinplatesplines import ThinPlateSplines
from .piecewiseaffine import PiecewiseAffine
from .rbf import R2LogR2RBF, R2LogRRBF
from .groupalign.procrustes import GeneralizedProcrustesAnalysis
from .compositions import (scale_about_centre, rotate_ccw_about_centre,
                           shear_about_centre, transform_about_centre)
from .tcoords import image_coords_to_tcoords, tcoords_to_image_coords


class WithDims(Transform):
    r"""
    Slices off select dimensions of a shape.

    Parameters
    ----------
    dims : valid numpy array slice
        The slice that will be used on the dimensionality axis of the shape
        under transform. For example, to go from a 3D shape to a 2D one,
        [0, 1] could be provided or np.array([True, True, False]).
    """
    def __init__(self, dims):
        self.dims = dims

    def _apply(self, x, **kwargs):
        # if self.dims is a single number we will return an array with the
        # spatial dimension missing - always reshape to avoid this case.
        return x[:, self.dims].reshape([x.shape[0], -1]).copy()



import numpy as np
from menpo.image import *


def test_blank_1_channel_image():
    mask = np.zeros((10, 10), dtype=np.bool)
    im = MaskedImage.init_blank((10, 10), mask=mask)
    assert np.all(im.pixels == 0.0)
    assert im.n_channels == 1
    assert np.all(im.mask.pixels == 0.0)

    im = MaskedImage.init_blank((10, 10), fill=0.5)
    assert np.all(im.pixels == 0.5)


def test_blank_3_channel_image():
    mask = np.zeros((10, 10), dtype=np.bool)
    im = MaskedImage.init_blank((10, 10), mask=mask, n_channels=3)
    assert np.all(im.pixels == 0.0)
    assert im.n_channels == 3
    assert np.all(im.mask.pixels == 0.0)

    im = MaskedImage.init_blank((10, 10), fill=0.5, n_channels=3)
    assert np.all(im.pixels == 0.5)


def test_blank_maskedimage():
    mask = np.zeros((10, 10), dtype=np.bool)
    im = MaskedImage.init_blank((10, 10), mask=mask, n_channels=10)
    assert np.all(im.pixels == 0.0)
    assert im.n_channels == 10
    assert np.all(im.mask.pixels == 0.0)

    im = MaskedImage.init_blank((10, 10), fill=2.0, n_channels=10)
    assert np.all(im.pixels == 2.0)



import numpy as np
from numpy.testing import assert_allclose
from pytest import raises
from scipy.sparse import csr_matrix, lil_matrix
from menpo.image import Image, MaskedImage
from menpo.shape import (UndirectedGraph, DirectedGraph, Tree,
                         PointUndirectedGraph, PointDirectedGraph, PointTree)

# Points for point graphs
points = np.array([[10, 30], [0, 20], [20, 20], [0, 10], [20, 10], [0, 0]])
points2 = np.array([[30, 30], [10, 20], [50, 20], [0, 10], [20, 10],
                    [50, 10], [0, 0], [20, 0], [50, 0]])
point = np.array([[10, 10]])

# Define undirected graph and pointgraph
adj_undirected = np.array([[0, 1, 1, 0, 0, 0],
                           [1, 0, 1, 1, 0, 0],
                           [1, 1, 0, 0, 1, 0],
                           [0, 1, 0, 0, 1, 1],
                           [0, 0, 1, 1, 0, 0],
                           [0, 0, 0, 1, 0, 0]])
g_undirected = UndirectedGraph(adj_undirected)
pg_undirected = PointUndirectedGraph(points, adj_undirected)

# Define directed graph and pointgraph
adj_directed = csr_matrix(([1] * 8, ([1, 2, 1, 2, 1, 2, 3, 3],
                                     [0, 0, 2, 1, 3, 4, 4, 5])), shape=(6, 6))
g_directed = DirectedGraph(adj_directed)
pg_directed = PointDirectedGraph(points, adj_directed)

# Define tree and pointtree
adj_tree = np.array([[0, 1, 1, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 1, 1, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 1, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 1, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 1, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 1],
                     [0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 0]])
g_tree = Tree(adj_tree, 0)
pg_tree = PointTree(points2, adj_tree, 0)

# Define undirected graph and pointgraph with isolated vertices
adj_isolated = csr_matrix(([1] * 6, ([0, 2, 2, 4, 3, 4], [2, 0, 4, 2, 4, 3])),
                          shape=(6, 6))
g_isolated = UndirectedGraph(adj_isolated)
pg_isolated = PointUndirectedGraph(points, adj_isolated)

# Define undirected graph and pointgraph with a single vertex
adj_single = np.array([[0]])
g_single = DirectedGraph(adj_single)
pg_single = PointDirectedGraph(point, adj_single)


def test_create_graph_exception():
    adj = np.array([[0, 1, 3],
                    [0, 2, 4]])
    with raises(ValueError):
        UndirectedGraph(adj)


def test_create_tree_exception_1():
    adj = csr_matrix(([1] * 9, ([0, 0, 1, 1, 2, 3, 4, 5, 0],
                                [1, 2, 3, 4, 5, 6, 7, 8, 4])), shape=(9, 9))
    with raises(ValueError):
        Tree(adj, 0)


def test_create_tree_exception_2():
    with raises(ValueError):
        Tree(adj_tree, 20)


def test_create_tree_exception_3():
    adj = csr_matrix(([1] * 8, ([0, 0, 1, 1, 2, 3, 4, 5],
                                [1, 2, 3, 4, 5, 6, 7, 8])), shape=(10, 10))
    with raises(ValueError):
        Tree(adj, 0)


def test_init_from_edges():
    g = PointDirectedGraph.init_from_edges(
        points, np.array([[1, 0], [2, 0], [1, 2], [2, 1], [1, 3], [2, 4],
                          [3, 4], [3, 5]]))
    assert (pg_directed.adjacency_matrix - g.adjacency_matrix).nnz == 0
    g = PointUndirectedGraph.init_from_edges(
        points, np.array([[0, 1], [0, 2], [1, 2], [1, 3], [2, 4], [3, 4],
                          [3, 5]]))
    assert (pg_undirected.adjacency_matrix - g.adjacency_matrix).nnz == 0
    g = PointUndirectedGraph.init_from_edges(
        points, np.array([[0, 1], [1, 0], [0, 2], [2, 0], [1, 2], [2, 1],
                          [1, 3], [3, 1], [2, 4], [4, 2], [3, 4], [4, 3],
                          [3, 5], [5, 3]]))
    assert (pg_undirected.adjacency_matrix - g.adjacency_matrix).nnz == 0
    g = PointTree.init_from_edges(
        points2, np.array([[0, 1], [0, 2], [1, 3], [1, 4], [2, 5], [3, 6],
                           [4, 7], [5, 8]]), root_vertex=0)
    assert (pg_tree.adjacency_matrix - g.adjacency_matrix).nnz == 0
    g = PointUndirectedGraph.init_from_edges(
        points, np.array([[0, 2], [2, 4], [3, 4]]))
    assert (pg_isolated.adjacency_matrix - g.adjacency_matrix).nnz == 0
    g = PointDirectedGraph.init_from_edges(point, np.array([]))
    assert (pg_single.adjacency_matrix - g.adjacency_matrix).nnz == 0


def test_init_2d_grid():
    g = PointTree.init_2d_grid((5, 5))
    assert g.adjacency_matrix.nnz == 24
    assert g.n_points == 25
    g = PointUndirectedGraph.init_2d_grid((5, 5))
    assert g.adjacency_matrix.nnz == 80
    assert g.n_points == 25
    g = PointDirectedGraph.init_2d_grid((5, 5))
    assert g.adjacency_matrix.nnz == 80
    assert g.n_points == 25


def test_init_2d_grid_custom_adjacency():
    tree_adj = np.array([[0, 1, 0, 0],
                         [0, 0, 0, 1],
                         [0, 0, 0, 0],
                         [0, 0, 1, 0]])
    g = PointTree.init_2d_grid((2, 2), root_vertex=0, adjacency_matrix=tree_adj)
    assert g.adjacency_matrix.nnz == 3
    assert g.n_points == 4
    single_edge = lil_matrix((25, 25))
    single_edge[[0, 1], [1, 0]] = 1
    single_edge = single_edge.tocsr()
    g = PointUndirectedGraph.init_2d_grid(
        (5, 5), adjacency_matrix=single_edge)
    assert g.adjacency_matrix.nnz == 2
    assert g.n_points == 25
    g = PointDirectedGraph.init_2d_grid(
        (5, 5), adjacency_matrix=single_edge)
    assert g.adjacency_matrix.nnz == 2
    assert g.n_points == 25


def test_init_from_depth_image():
    fake_z = np.random.uniform(size=(10, 10))
    g = PointTree.init_from_depth_image(Image(fake_z))
    assert g.n_dims == 3
    assert g.root_vertex == 55
    assert g.adjacency_matrix.nnz == 99
    assert g.n_points == 100
    g = PointUndirectedGraph.init_from_depth_image(Image(fake_z))
    assert g.n_dims == 3
    assert g.adjacency_matrix.nnz == 360
    assert g.n_points == 100
    g = PointDirectedGraph.init_from_depth_image(Image(fake_z))
    assert g.n_dims == 3
    assert g.adjacency_matrix.nnz == 360
    assert g.n_points == 100


def test_init_from_depth_image_masked():
    fake_z = np.random.uniform(size=(10, 10))
    mask = np.zeros(fake_z.shape, dtype=np.bool)
    mask[2:6, 2:6] = True
    im = MaskedImage(fake_z, mask=mask)
    g = PointTree.init_from_depth_image(im)
    assert g.n_dims == 3
    assert g.root_vertex == 0
    assert g.adjacency_matrix.nnz == 15
    assert g.n_points == 16
    g = PointUndirectedGraph.init_from_depth_image(im)
    assert g.n_dims == 3
    assert g.adjacency_matrix.nnz == 48
    assert g.n_points == 16
    g = PointDirectedGraph.init_from_depth_image(im)
    assert g.n_dims == 3
    assert g.adjacency_matrix.nnz == 48
    assert g.n_points == 16


def test_n_edges():
    assert_allclose(g_directed.n_edges, 8)
    assert_allclose(g_undirected.n_edges, 7)
    assert_allclose(g_tree.n_edges, 8)
    assert_allclose(g_isolated.n_edges, 3)
    assert_allclose(g_single.n_edges, 0)


def test_edges():
    assert_allclose(g_directed.edges, np.array([[1, 0], [1, 2], [1, 3], [2, 0],
                                                [2, 1], [2, 4], [3, 4],
                                                [3, 5]]))
    assert_allclose(g_undirected.edges, np.array([[0, 1], [0, 2], [1, 2],
                                                  [1, 3], [2, 4], [3, 4],
                                                  [3, 5]]))
    assert_allclose(g_tree.edges, np.array([[0, 1], [0, 2], [1, 3], [1, 4],
                                            [2, 5], [3, 6], [4, 7], [5, 8]]))
    assert_allclose(g_isolated.edges, np.array([[0, 2], [2, 4], [3, 4]]))
    assert_allclose(g_single.edges, np.empty((0, 2)))


def test_n_vertices():
    assert_allclose(g_directed.n_vertices, 6)
    assert_allclose(g_undirected.n_vertices, 6)
    assert_allclose(g_tree.n_vertices, 9)
    assert_allclose(g_isolated.n_vertices, 6)
    assert_allclose(g_single.n_vertices, 1)


def test_vertices():
    assert (list(g_directed.vertices) == [0, 1, 2, 3, 4, 5])
    assert (list(g_undirected.vertices) == [0, 1, 2, 3, 4, 5])
    assert (list(g_tree.vertices) == [0, 1, 2, 3, 4, 5, 6, 7, 8])
    assert (list(g_isolated.vertices) == [0, 1, 2, 3, 4, 5])
    assert (list(g_single.vertices) == [0])


def test_isolated_vertices():
    assert (g_directed.isolated_vertices() == [])
    assert not g_directed.has_isolated_vertices()
    assert (g_undirected.isolated_vertices() == [])
    assert not g_undirected.has_isolated_vertices()
    assert (g_tree.isolated_vertices() == [])
    assert not g_tree.has_isolated_vertices()
    assert (g_isolated.isolated_vertices() == [1, 5])
    assert g_isolated.has_isolated_vertices()
    assert (g_single.isolated_vertices() == [0])
    assert g_single.has_isolated_vertices()


def test_adjacency_list():
    assert (g_directed.get_adjacency_list() ==
            [[], [0, 2, 3], [0, 1, 4], [4, 5], [], []])
    assert (g_undirected.get_adjacency_list() ==
            [[1, 2], [0, 2, 3], [0, 1, 4], [1, 4, 5], [2, 3], [3]])
    assert (g_tree.get_adjacency_list() ==
            [[1, 2], [3, 4], [5], [6], [7], [8], [], [], []])
    assert (g_isolated.get_adjacency_list() ==
            [[2], [], [0, 4], [4], [2, 3], []])
    assert (g_single.get_adjacency_list() == [[]])


def test_n_paths():
    assert_allclose(g_directed.n_paths(1, 4), 2)
    assert_allclose(g_undirected.n_paths(0, 5), 4)
    assert_allclose(g_tree.n_paths(2, 6), 0)
    assert_allclose(g_isolated.n_paths(3, 2), 1)
    assert_allclose(g_single.n_paths(0, 0), 1)


def test_find_all_paths():
    assert (g_directed.find_all_paths(1, 4) == [[1, 2, 4], [1, 3, 4]])
    assert (g_undirected.find_all_paths(0, 5) ==
            [[0, 1, 2, 4, 3, 5], [0, 1, 3, 5], [0, 2, 1, 3, 5],
             [0, 2, 4, 3, 5]])
    assert (g_tree.find_all_paths(2, 6) == [])
    assert (g_isolated.find_all_paths(1, 5) == [])
    assert (g_single.find_all_paths(0, 10) == [])


def test_find_path():
    assert (g_directed.find_all_paths(1, 4)[0] == g_directed.find_path(1, 4))
    assert (g_undirected.find_all_paths(0, 5)[1] ==
            g_undirected.find_path(0, 5))
    assert (g_tree.find_path(2, 6) == [])
    assert (g_isolated.find_path(4, 1) == [])
    assert (g_single.find_path(0, 0) == [])


def test_find_shortest_path():
    assert (g_directed.find_shortest_path(1, 0) == ([1, 0], 0.0))
    assert (g_undirected.find_shortest_path(5, 0) == ([5, 3, 1, 0], 3.0))
    assert (g_tree.find_shortest_path(1, 7) == ([1, 4, 7], 1.0))
    assert (g_isolated.find_shortest_path(3, 0) == ([3, 4, 2, 0], 3.0))
    assert (g_single.find_shortest_path(0, 0) == ([], np.inf))


def test_neighbours_children_parent():
    assert (g_directed.children(3) == [4, 5])
    assert (g_directed.n_children(0) == 0)
    assert (g_directed.parents(0) == [1, 2])
    assert (g_directed.n_parents(3) == 1)

    assert (g_undirected.neighbours(1) == [0, 2, 3])
    assert (g_undirected.n_neighbours(3) == 3)

    assert (g_tree.children(5) == [8])
    assert (g_tree.n_children(6) == 0)
    assert (g_tree.parent(4) == [1])

    assert (g_isolated.neighbours(1) == [])
    assert (g_isolated.n_neighbours(4) == 2)

    assert (g_single.children(0) == [])
    assert (g_single.n_children(0) == 0)
    assert (g_single.parents(0) == [])
    assert (g_single.n_parents(0) == 0)


def test_tree():
    assert (g_tree.maximum_depth == 3)
    assert (g_tree.depth_of_vertex(0) == 0)
    assert (g_tree.depth_of_vertex(4) == 2)
    assert (g_tree.n_vertices_at_depth(2) == 3)
    assert g_tree.is_leaf(7)
    assert not g_tree.is_leaf(5)
    assert (g_tree.leaves == [6, 7, 8])
    assert (g_tree.n_leaves == 3)


def test_minimum_spanning_tree():
    adjacency_matrix = np.array([[0, 11, 13, 12],
                                 [11, 0, 0, 14],
                                 [13, 0, 0, 10],
                                 [12, 14, 10, 0]])
    g = UndirectedGraph(adjacency_matrix)
    t = g.minimum_spanning_tree(root_vertex=0)
    assert t.n_edges == 3
    assert_allclose(t.adjacency_matrix.todense(),
                    csr_matrix(([11., 12., 10.], ([0, 0, 3], [1, 3, 2])),
                               shape=(4, 4)).todense())
    assert t.get_adjacency_list() == [[1, 3], [], [], [2]]
    assert t.predecessors_list == [None, 0, 3, 0]


def test_is_edge():
    assert g_directed.is_edge(2, 4)
    assert not g_directed.is_edge(3, 1)
    assert not g_directed.is_edge(5, 0)

    assert g_undirected.is_edge(2, 1)
    assert g_undirected.is_edge(1, 2)
    assert not g_undirected.is_edge(5, 0)

    assert g_tree.is_edge(4, 7)
    assert not g_tree.is_edge(6, 3)
    assert not g_tree.is_edge(8, 1)

    assert g_isolated.is_edge(0, 2)
    assert g_isolated.is_edge(4, 2)
    assert not g_isolated.is_edge(1, 5)

    assert not g_single.is_edge(0, 0)


def test_is_tree():
    assert not g_undirected.is_tree()
    assert not g_directed.is_tree()
    assert g_tree.is_tree()
    assert not g_isolated.is_tree()
    assert g_single.is_tree()


def test_from_mask():
    assert (pg_directed.from_mask(
        np.array([False, False, True,
                  True, True, False])).get_adjacency_list() == [[2], [2], []])
    assert (pg_undirected.from_mask(
        np.array([True, False, True,
                  True, True, False])).get_adjacency_list() ==
            [[1], [0, 3], [3], [1, 2]])
    assert (pg_tree.from_mask(
        np.array([True, False, True, True, True,
                  True, True, True, False])).get_adjacency_list() ==
            [[1], [2], []])
    assert (pg_isolated.from_mask(
        np.array([True, True, False,
                  True, True, False])).get_adjacency_list() ==
            [[], [], [3], [2]])
    assert (pg_single.from_mask(np.array([True])).get_adjacency_list() == [[]])


def test_from_mask_errors():
    with raises(ValueError):
        pg_directed.from_mask(np.array([False, False, False, False, False,
                                        False]))
    with raises(ValueError):
        pg_undirected.from_mask(np.array([False, False, False, False, False,
                                          False]))
    with raises(ValueError):
        pg_tree.from_mask(np.array([False, True, True, True, True, True, True,
                                    True, True]))
    with raises(ValueError):
        pg_isolated.from_mask(np.array([False, False, False, False, False,
                                        False]))
    with raises(ValueError):
        pg_single.from_mask(np.array([False]))


def test_relative_locations():
    assert_allclose(pg_tree.relative_location_edge(5, 8), np.array([0, -10]))
    assert_allclose(pg_tree.relative_locations(), np.array([[-20, -10],
                                                            [20, -10],
                                                            [-10, -10],
                                                            [10, -10],
                                                            [0, -10],
                                                            [0, -10],
                                                            [0, -10],
                                                            [0, -10]]))


def test_relative_locations():
    with raises(ValueError):
        pg_tree.relative_location_edge(8, 5)
    with raises(ValueError):
        pg_tree.relative_location_edge(0, 6)



import numpy as np
from mock import patch, MagicMock
from pytest import raises
from scipy.sparse import csr_matrix

from menpo.image import Image
from menpo.visualize import MenpowidgetsMissingError
from menpo.shape import PointCloud, TriMesh, PointDirectedGraph, LabelledPointUndirectedGraph
from menpo.landmark import LandmarkManager
from menpo.testing import surrogate

pcloud2d = np.zeros([1, 2])
triangle_pcloud2d = np.array([[0, 0], [0, 1], [1, 0.]])
menpowidgets_mock = MagicMock()
menpowidgets_mock.side_effect = ImportError


@surrogate('menpowidgets.visualize_images')
@patch('menpowidgets.visualize_images', menpowidgets_mock)
def test_image_view_widget():
    with raises(MenpowidgetsMissingError):
        Image.init_blank((5, 5)).view_widget()


@surrogate('menpowidgets.visualize_pointclouds')
@patch('menpowidgets.visualize_pointclouds', menpowidgets_mock)
def test_pointcloud_view_widget():
    with raises(MenpowidgetsMissingError):
        PointCloud(pcloud2d).view_widget()


@surrogate('menpowidgets.visualize_landmarkgroups')
@patch('menpowidgets.visualize_landmarkgroups', menpowidgets_mock)
def test_labelledpointundirectedgraph_view_widget():
    adj_matrix = csr_matrix((1, 1))
    with raises(MenpowidgetsMissingError):
        LabelledPointUndirectedGraph.init_with_all_label(
            pcloud2d, adj_matrix).view_widget()


@surrogate('menpowidgets.visualize_landmarks')
@patch('menpowidgets.visualize_landmarks', menpowidgets_mock)
def test_landmarkmanager_view_widget():
    l = LandmarkManager()
    l['test'] = PointCloud(pcloud2d)
    with raises(MenpowidgetsMissingError):
        l.view_widget()


@surrogate('menpowidgets.visualize_patches')
@patch('menpowidgets.visualize_patches', menpowidgets_mock)
def test_landmarkmanager_view_widget():
    l = LandmarkManager()
    l['test'] = PointCloud(pcloud2d)
    with raises(MenpowidgetsMissingError):
        l.view_widget()


@surrogate('menpowidgets.visualize_pointclouds')
@patch('menpowidgets.visualize_pointclouds', menpowidgets_mock)
def test_trimesh_view_widget():
    with raises(MenpowidgetsMissingError):
        TriMesh(triangle_pcloud2d).view_widget()


@surrogate('menpowidgets.visualize_pointclouds')
@patch('menpowidgets.visualize_pointclouds', menpowidgets_mock)
def test_pointgraph_view_widget():
    with raises(MenpowidgetsMissingError):
        PointDirectedGraph.init_from_edges(triangle_pcloud2d,
                                           [[0, 1], [1, 2]]).view_widget()



import pytest

import menpo.landmark.labels as labels
import inspect
import numpy as np
from menpo.shape import PointCloud


EXEMPT_FUNCTIONS = {'labeller', 'bounding_box_to_bounding_box',
                    'bounding_box_mirrored_to_bounding_box'}

valid_label_functions = filter(
    lambda x: x[0] not in EXEMPT_FUNCTIONS,
    inspect.getmembers(labels, predicate=inspect.isfunction)
)


def check_label_func(func, input_n_points, output_n_points):
    # Could be any dimensionality
    array = np.zeros([input_n_points, 2])
    pcloud = PointCloud(array)

    array_result = func(array)
    assert isinstance(array_result, PointCloud)
    assert array_result.n_points == output_n_points

    pcloud_result = func(pcloud)
    assert isinstance(pcloud_result, PointCloud)
    assert pcloud_result.n_points == output_n_points


@pytest.mark.parametrize('fname, func', valid_label_functions)
def test_labels(fname, func):
    parts = fname.split('_')
    to_index = parts.index('to')
    input_n_points = parts[to_index - 1]
    output_n_points = parts[-1]
    try:
        input_n_points = int(input_n_points)
    except ValueError:
        # If it isn't the element before 'to', there must be
        # a modifier - so try the element before that
        input_n_points = int(parts[to_index - 2])
    try:
        output_n_points = int(output_n_points)
    except ValueError:
        # If it isn't the last element, there must be
        # a modifier - so try the second to last element
        output_n_points = int(parts[-2])

    check_label_func(func, input_n_points, output_n_points)


def test_bounding_box_to_bounding_box():
    from menpo.landmark import bounding_box_to_bounding_box

    check_label_func(bounding_box_to_bounding_box, 4, 4)


def test_bounding_box_mirrored_to_bounding_box():
    from menpo.landmark import bounding_box_mirrored_to_bounding_box

    check_label_func(bounding_box_mirrored_to_bounding_box, 4, 4)



import gzip
import warnings
from functools import partial
from pathlib import Path

from menpo.compatibility import basestring, str
from .extensions import landmark_types, image_types, pickle_types, video_types
from ..exceptions import OverwriteError
from ..utils import (_norm_path, _possible_extensions_from_filepath,
                     _normalize_extension)

# an open file handle that uses a small fast level of compression
gzip_open = partial(gzip.open, compresslevel=3)


def export_landmark_file(landmarks_object, fp, extension=None, overwrite=False):
    r"""
    Exports a given shape. The ``fp`` argument can be either or a `str` or
    any Python type that acts like a file. If a file is provided, the
    ``extension`` kwarg **must** be provided. If no ``extension`` is provided
    and a `str` filepath is provided, then the export type is calculated
    based on the filepath extension.

    Due to the mix in string and file types, an explicit overwrite argument is
    used which is ``False`` by default.

    Parameters
    ----------
    landmarks_object : dict or :map:`LandmarkManager`  or
        :map:`PointCloud` or subclass of :map:`PointCloud`
        The landmarks to export. The type of :map:`PointCloud` or
        subclass of it are supported by all exporters, while the
        rest are available only for the LJSON format.
    fp : `Path` or `file`-like object
        The Path or file-like object to save the object at/into.
    extension : `str` or None, optional
        The extension to use, this must match the file path if the file
        path is a string. Determines the type of exporter that is used.
    overwrite : `bool`, optional
        Whether or not to overwrite a file if it already exists.

    Raises
    ------
    ValueError
        File already exists and ``overwrite`` != ``True``
    ValueError
        ``fp`` is a `str` and the ``extension`` is not ``None``
        and the two extensions do not match
    ValueError
        ``fp`` is a `file`-like object and ``extension`` is
        ``None``
    ValueError
        The provided extension does not match to an existing exporter type
        (the output type is not supported).
    ValueError
        The provided type for landmarks_object is not supported.
    """
    extension = _normalize_extension(extension)

    try:
        landmarks_object.n_points
    except AttributeError:
        # unless this is LJSON, this is not correct.
        fp_is_path = isinstance(fp, basestring) or isinstance(fp, Path)
        if (extension is not None and extension != '.ljson') or \
                (fp_is_path and Path(fp).suffix != '.ljson'):
            m1 = ('Only the LJSON format supports multiple '
                  'keys for exporting. \nIn any other '
                  'case your input should be a PointCloud or '
                  'subclass.')
            raise ValueError(m1)
    _export(landmarks_object, fp, landmark_types, extension, overwrite)


def export_image(image, fp, extension=None, overwrite=False):
    r"""
    Exports a given image. The ``fp`` argument can be either
    a `Path` or any Python type that acts like a file. If a file is provided,
    the ``extension`` kwarg **must** be provided. If no
    ``extension`` is provided and a `str` filepath is provided, then
    the export type is calculated based on the filepath extension.

    Due to the mix of string and file types, an explicit overwrite argument is
    used which is ``False`` by default.

    Parameters
    ----------
    image : :map:`Image`
        The image to export.
    fp : `Path` or `file`-like object
        The Path or file-like object to save the object at/into.
    extension : `str` or None, optional
        The extension to use, this must match the file path if the file
        path is a string. Determines the type of exporter that is used.
    overwrite : `bool`, optional
        Whether or not to overwrite a file if it already exists.

    Raises
    ------
    ValueError
        File already exists and ``overwrite`` != ``True``
    ValueError
        ``fp`` is a `str` and the ``extension`` is not ``None``
        and the two extensions do not match
    ValueError
        ``fp`` is a `file`-like object and ``extension`` is
        ``None``
    ValueError
        The provided extension does not match to an existing exporter type
        (the output type is not supported).
    """
    _export(image, fp, image_types, extension, overwrite)


def export_video(images, file_path, overwrite=False, fps=30, **kwargs):
    r"""
    Exports a given list of images as a video. Ensure that all the images
    have the same shape, otherwise you might get unexpected results from
    the ffmpeg writer. The ``file_path`` argument is a `Path` representing
    the path to save the video to. At this time, it is not possible
    to export videos directly to a file buffer.

    Due to the mix of string and file types, an explicit overwrite argument is
    used which is ``False`` by default.

    Note that exporting of GIF images is also supported.

    Parameters
    ----------
    images : list of :map:`Image`
        The images to export as a video.
    file_path : `Path`
        The Path to save the video at. File buffers are not supported, unlike
        other exporting formats.
    overwrite : `bool`, optional
        Whether or not to overwrite a file if it already exists.
    fps : `int`, optional
        The number of frames per second.
    **kwargs : `dict`, optional
        Extra parameters that are passed through directly to the exporter.
        Please see the documentation in the ``menpo.io.output.video`` package
        for information about the supported arguments.

    Raises
    ------
    ValueError
        File already exists and ``overwrite`` != ``True``
    ValueError
        The input is a buffer and not a valid `Path`
    ValueError
        The provided extension does not match to an existing exporter type
        (the output type is not supported).
    """
    exporter_kwargs = {'fps': fps}
    exporter_kwargs.update(kwargs)

    file_path = _enforce_only_paths_supported(file_path, 'FFMPEG')
    _export_paths_only(images, file_path, video_types, None, overwrite,
                       exporter_kwargs=exporter_kwargs)


def export_pickle(obj, fp, overwrite=False, protocol=2):
    r"""
    Exports a given collection of Python objects with Pickle.

    The ``fp`` argument can be either a `Path` or any Python type that acts like
    a file.
    If ``fp`` is a path, it must have the suffix `.pkl` or `.pkl.gz`. If
    `.pkl`, the object will be pickled using the selected Pickle protocol.
    If `.pkl.gz` the object will be pickled using the selected Pickle
    protocol with gzip compression (at a fixed compression level of 3).

    Note that a special exception is made for `pathlib.Path` objects - they
    are pickled down as a `pathlib.PurePath` so that pickles can be easily
    moved between different platforms.

    Parameters
    ----------
    obj : ``object``
        The object to export.
    fp : `Path` or `file`-like object
        The string path or file-like object to save the object at/into.
    overwrite : `bool`, optional
        Whether or not to overwrite a file if it already exists.
    protocol : `int`, optional
        The Pickle protocol used to serialize the file.
        The protocols were introduced in different versions of python, thus
        it is recommended to save with the highest protocol version that
        your python distribution can support.
        The protocol refers to:

        ========= =========================================================
        Protocol                       Functionality
        ========= =========================================================
        0         Simplest protocol for text mode, backwards compatible.
        1         Protocol for binary mode, backwards compatible.
        2         Wider support for classes, compatible with python >= 2.3.
        3         Support for byte objects, compatible with python >= 3.0.
        4         Support for large objects, compatible with python >= 3.4.
        ========= =========================================================
    Raises
    ------
    ValueError
        File already exists and ``overwrite`` != ``True``
    ValueError
        ``fp`` is a `file`-like object and ``extension`` is
        ``None``
    ValueError
        The provided extension does not match to an existing exporter type
        (the output type is not supported).
    """
    exporter_kwargs = {'protocol': protocol}
    if isinstance(fp, basestring):
        fp = Path(fp)  # cheeky conversion to Path to reuse existing code
    if isinstance(fp, Path):
        # user provided a path - if it ended .gz we will compress
        path_filepath = _validate_filepath(fp, overwrite)
        extension = _parse_and_validate_extension(path_filepath, None,
                                                  pickle_types)
        o = gzip_open if extension[-3:] == '.gz' else open
        with o(str(path_filepath), 'wb') as f:
            # force overwrite as True we've already done the check above
            _export(obj, f, pickle_types, extension, True,
                    exporter_kwargs=exporter_kwargs)
    else:
        _export(obj, fp, pickle_types, '.pkl', overwrite,
                exporter_kwargs=exporter_kwargs)


def _extension_to_export_function(extension, extensions_map):
    r"""
    Simple function that wraps the extensions map indexing and raises
    a user friendly ``ValueError``

    Parameters
    ----------
    extension : `str`
        The string extension with period prefix e.g '.jpg'
    extensions_map : `dict` of `str` -> `callable`
        The extension map that maps extensions to export callables.

    Returns
    -------
    mapping_callable : `callable`
        The callable that performs exporting.

    Raises
    ------
    ValueError
        If ``extensions_map`` does not contain ``extension``. More friendly
        than the ``KeyError`` that would be raised.
    """
    # This just ensures that a sensible, user friendly Exception is raised.
    try:
        return extensions_map[extension]
    except KeyError:
        raise ValueError('The output file extension ({}) provided is not '
                         'currently supported.'.format(extension))


def _validate_filepath(fp, overwrite):
    r"""
    Normalize a given file path and ensure that ``overwrite == True`` if the
    file path exists. Normalisation involves things like making the given
    path absolute and expanding environment variables and user variables.

    Parameters
    ----------
    fp : `Path`
        The file path.
    overwrite : `bool`
        Whether the export method should override an existing file at the
        file path.

    Returns
    -------
    normalized_filepath : `Path`
        The normalized file path.

    Raises
    ------
    OverwriteError
        If ``overwrite == False`` and a file already exists at the file path.
    """
    path_filepath = _norm_path(fp)
    if path_filepath.exists() and not overwrite:
        raise OverwriteError('File {} already exists. Please set the overwrite '
                             'kwarg if you wish to overwrite '
                             'the file.'.format(path_filepath.name),
                             path_filepath)
    return path_filepath


def _parse_and_validate_extension(filepath, extension, extensions_map):
    r"""
    If an extension is given, validate that the given file path matches
    the given extension.

    If not, parse the file path and return a correct extension. This function
    will handle cases such as file names with periods in.

    Parameters
    ----------
    filepath : `Path`
        The file path (normalized).
    extension : `str`
        The extension provided by the user.
    extensions_map : `dict` of `str` -> `callable`
        A dictionary mapping extensions to export callables.

    Returns
    -------
    norm_extension : `str`
        The correct extension, with leading period.

    Raises
    ------
    ValueError
        Unknown extension.
    ValueError
        File path contains extension that does not EXACTLY match the users'
        provided extension.
    """
    # If an explicit extension is passed, it must match exactly. However, file
    # names may contain periods, and therefore we need to try and parse
    # a known extension from the given file path.
    possible_exts = _possible_extensions_from_filepath(filepath)

    known_extension = None
    while known_extension is None and possible_exts:
        possible_extension = possible_exts.pop(0)
        if possible_extension in extensions_map:
            known_extension = possible_extension

    if known_extension is None:
        raise ValueError('Unknown file extension passed: {}'.format(
            ''.join(filepath.suffixes)))

    if extension is not None:
        extension = _normalize_extension(extension)
        if extension != known_extension:
            raise ValueError('The file path extension must match the '
                             'requested file extension: {} != {}'.format(
                               extension, known_extension))

    return known_extension


def _enforce_only_paths_supported(file_path, exporter_name):
    r"""
    If a given exporter only supports paths rather than open file handles
    or buffers then this function can be used to enforce that. If a file
    handle is passed then an attempt is made to write to the path of the file
    handle.

    Parameters
    ----------
    file_path : `str` or `pathlib.Path` or file-like object
        The file path to write to.

    Returns
    -------
    file_path : `str`
        The path to open file handle or, if a path was passed, it is returned
        unchanged.

    Raises
    ------
    ValueError
        If given ``file_path`` is not a string, pathlib.Path or file handle.
    """
    if hasattr(file_path, 'name') and not isinstance(file_path, Path):
        file_path = file_path.name
        warnings.warn('The {} exporter only supports file paths and not '
                      'buffers or open file handles - therefore the provided '
                      'file handle will be ignored and the object will be '
                      'exported to {}.'.format(exporter_name, file_path))
    if isinstance(file_path, basestring) or isinstance(file_path, Path):
        return file_path
    else:
        raise ValueError('Cannot write to unnamed file handles or buffers.')


def _validate_and_get_export_func(file_path, extensions_map, extension,
                                  overwrite, return_extension=False):
    r"""
    Given a ``file_path``, ensure that the options chosen are valid with respect
    to overwriting and any provided extensions. If this validation is
    successful then the exporter function is returned.

    Parameters
    ----------
    file_path : `Path`
        The path to write to.
    extensions_map : `dict` of `str` -> `callable`
        The dictionary mapping extensions to export callables.
    extension : `str`
        User provided extension (required if a file-like ``fp`` is passed).
    overwrite : `bool`
        If ``True``, overwrite any existing files at the given path.
    return_extension : `bool`, optional
        If ``True``, return the correct extension as well as the export
        callable, as a tuple ``(callable, extension)``.

    Returns
    -------
    exporter_callable : `callable`
        The exporter callable.
    extension : `str`
        The correct extension for the exporter function, if
        ``return_extension==True``.
    """
    if isinstance(file_path, basestring):
        # cheeky conversion to Path to reuse existing code
        file_path = Path(file_path)

    file_path = _validate_filepath(file_path, overwrite)
    extension = _parse_and_validate_extension(file_path, extension,
                                              extensions_map)
    export_callable = _extension_to_export_function(extension, extensions_map)

    if return_extension:
        return export_callable, extension
    else:
        return export_callable


def _export_paths_only(obj, file_path, extensions_map, extension, overwrite,
                       exporter_kwargs=None):
    r"""
    A shared export function handling paths only. This handles the logic
    of ensuring that the given ``file_path`` is a ``pathlib.Path``. All exporter
    methods that are called from here are defined as receiving a
    ``pathlib.Path``.

    Parameters
    ----------
    obj : `object`
        The Python object to export.
    file_path : `Path`
        The path to write to.
    extensions_map : `dict` of `str` -> `callable`
        The dictionary mapping extensions to export callables.
    extension : `str`
        User provided extension (required if a file-like ``fp`` is passed).
    overwrite : `bool`
        If ``True``, overwrite any existing files at the given path.
    exporter_kwargs : `int`, optional
        Any kwargs to be passed through to the exporter.
    """
    if exporter_kwargs is None:
        exporter_kwargs = {}
    export_function = _validate_and_get_export_func(file_path, extensions_map,
                                                    extension, overwrite)
    export_function(obj, file_path, **exporter_kwargs)


def _export(obj, fp, extensions_map, extension, overwrite,
            exporter_kwargs=None):
    r"""
    The shared export function. This handles the shared logic of ensuring
    that the given ``fp`` is either a ``pathlib.Path`` or a file like
    object. All exporter methods are defined as receiving a buffer object,
    regardless of if a path is provided. If a file-like object is provided
    then the extension mut not be ``None``.

    Parameters
    ----------
    obj : `object`
        The Python object to export.
    fp : `Path` or file-like object
        The path or file buffer to write to.
    extensions_map : `dict` of `str` -> `callable`
        The dictionary mapping extensions to export callables.
    extension : `str`
        User provided extension (required if a file-like ``fp`` is passed).
    overwrite : `bool`
        If ``True``, overwrite any existing files at the given path.
    exporter_kwargs : `int`, optional
        Any kwargs to be passed through to the exporter.
    """
    if exporter_kwargs is None:
        exporter_kwargs = {}
    if isinstance(fp, basestring):
        fp = Path(fp)  # cheeky conversion to Path to reuse existing code
    if isinstance(fp, Path):
        export_function, extension = _validate_and_get_export_func(
            fp, extensions_map, extension, overwrite, return_extension=True)

        with fp.open('wb') as file_handle:
            export_function(obj, file_handle, extension=extension,
                            **exporter_kwargs)
    else:
        # You MUST provide an extension if a file handle is given
        if extension is None:
            raise ValueError('An export file extension must be provided if a '
                             'file-like object is passed.')
        else:
            extension = _normalize_extension(extension)

        # Apparently in Python 2.x there is no reliable way to detect something
        # that is 'file' like (file handle or a StringIO object or something
        # you can read and write to like a file). Therefore, we are going to
        # just be really Pythonic about it and just assume we were handed
        # a correctly behaving object.
        try:
            # Follow PIL like behaviour. Check the file handle extension
            # and check if matches the given extension
            export_function = _validate_and_get_export_func(
                Path(fp.name), extensions_map, extension, overwrite)
        except AttributeError:
            # Just use the extension to get the export function
            export_function = _extension_to_export_function(extension,
                                                            extensions_map)

        export_function(obj, fp, extension=extension, **exporter_kwargs)



import numpy as np
from menpo.shape import PointCloud
from menpo.transform import WithDims


def test_withdims_transform():
    pc_3d = PointCloud(np.random.random((14, 3)))
    pc_2d = WithDims([0, 1]).apply(pc_3d)
    assert np.all(pc_3d.points[:, :2] == pc_2d.points)


def test_withdims_single_axis_transform():
    pc_3d = PointCloud(np.random.random((14, 3)))
    pc_1d = WithDims(0).apply(pc_3d)
    assert np.all(pc_3d.points[:, :0] == pc_1d.points)
    assert pc_1d.n_dims == 1
    assert pc_1d.n_points == pc_3d.n_points


def test_with_dims_method():
    pc_3d = PointCloud(np.random.random((14, 3)))
    pc_2d = pc_3d.with_dims([0, 1])
    assert np.all(pc_3d.points[:, :2] == pc_2d.points)


def test_with_dims_flip():
    pc_3d = PointCloud(np.random.random((14, 3)))
    pc_2d = pc_3d.with_dims([1, 0])
    assert np.all(pc_3d.points[:, :2] == pc_2d.points[:, ::-1])


def test_with_dims_repeat():
    pc_3d = PointCloud(np.random.random((14, 3)))
    pc_3d_rep = pc_3d.with_dims([0, 0, 1])
    assert np.all(pc_3d.points[:, 0] == pc_3d_rep.points[:, 0])
    assert np.all(pc_3d.points[:, 0] == pc_3d_rep.points[:, 1])
    assert np.all(pc_3d.points[:, 1] == pc_3d_rep.points[:, 2])



import numpy as np

from .interpolation import scipy_interpolation


def _centered_patch(patch_shape):
    r"""
    Helper method for generating the relative coordinates of a centered
    patch with shape (prod(patch_shape), 2). First generate a centered patch
    where the (0, 0) coordinate is in the center of the patch.
    Therefore, the (0, 0) pixel value is actually negative. The reshape to be in
    the form of sampling locations.

    Parameters
    ----------
    patch_shape : (R, C)

    Returns
    -------
    sampling_locations : ``(prod(patch_shape), 2)`` ``ndarray``
        The locations to sample for the given patch shape
    """
    assert len(patch_shape) == 2, 'Only 2D images are supported'
    half_pixel = (np.array([patch_shape]) % 2) / 2
    patch = np.meshgrid(
        np.linspace(-patch_shape[0] / 2, patch_shape[0] / 2, num=patch_shape[0],
                    dtype=np.float, endpoint=False),
        np.linspace(-patch_shape[1] / 2, patch_shape[1] / 2, num=patch_shape[1],
                    dtype=np.float, endpoint=False),
        indexing='ij'
    )
    patch = np.stack(patch, axis=2).reshape([-1, 2])
    return np.require(patch, requirements=['C']) + half_pixel


def extract_patches_by_sampling(pixels, patch_centers, patch_shape,
                                offsets=None, order=0, mode='constant',
                                cval=0.0):
    r"""
    Extract a set of patches from the given pixels. Given a set of patch centers
    and a patch size, patches are extracted from within the pixels, centered
    on the given coordinates. Note that only 2D images are currently supported.
    Note that this is much less efficient than slicing.

    Parameters
    ----------
    pixels : ``(n_channels, height, width)`` `ndarray``
        Pixels to extract patches from
    patch_centers : ``(n_patches, 2)`` `ndarray``
        The centers to extract patches around.
    patch_shape : ``(1, 2)`` `tuple` or `ndarray`, optional
        The size of the patch to extract
    offsets : ``(n_offsets, n_dims)`` `ndarray` or ``None``, optional
        Offsets to sample from within a patch. So ``(0, 0)`` is the
        centre of the patch (no offset) and ``(1, 0)`` would be sampling the
        patch from 1 pixel up the first axis away from the centre.
        If ``None``, then no offsets are applied.
    order : `int`, optional
        The order of interpolation. The order has to be in the range [0,5].
        See warp_to_shape for more information.
    mode : ``{constant, nearest, reflect, wrap}``, optional
        Points outside the boundaries of the input are filled according
        to the given mode.
    cval : `float`, optional
        Used in conjunction with mode ``constant``, the value outside
        the image boundaries.

    Returns
    -------
    patches : ``(n_center, n_offset, n_channels, patch_shape)`` `ndarray
        The extracted patches including any offsets if they were requested.

    Raises
    ------
    ValueError
        If pixels array is not 2D
    """
    if pixels.ndim != 3:
        raise ValueError('Only 2D images are supported but '
                         'found {}'.format(pixels.shape))

    n_points = patch_centers.shape[0]
    n_offsets = 1

    patch = _centered_patch(patch_shape)
    points_to_sample = patch[:, None, :] + patch_centers
    if offsets is not None:
        n_offsets = offsets.shape[0]
        points_to_sample = points_to_sample[:, :, None, :] + offsets
    points_to_sample = points_to_sample.reshape([-1, 2])

    patches = scipy_interpolation(pixels, points_to_sample,
                                  order=order, mode=mode, cval=cval)
    patches = patches.reshape(3, patch_shape[0], patch_shape[1],
                              n_points, n_offsets)
    patches = np.transpose(patches, [3, 4, 0, 1, 2])
    return np.require(patches, requirements=['C'])


def extract_patches_with_slice(pixels, patch_centers, patch_shape,
                               offsets=None, cval=0.):
    r"""
    Extract a set of patches from the given pixels. Given a set of patch centers
    and a patch size, patches are extracted from within the pixels, centered
    on the given coordinates. Note that only 2D images are currently supported.
    This is equivalent to sampling with ``order=0`` and ``mode='constant'``.

    Parameters
    ----------
    pixels : ``(n_channels, height, width)`` `ndarray``
        Pixels to extract patches from
    patch_centers : ``(n_patches, 2)`` `ndarray``
        The centers to extract patches around.
    patch_shape : ``(1, 2)`` `tuple` or `ndarray`, optional
        The size of the patch to extract
    offsets : ``(n_offsets, n_dims)`` `ndarray` or ``None``, optional
        Offsets to sample from within a patch. So ``(0, 0)`` is the
        centre of the patch (no offset) and ``(1, 0)`` would be sampling the
        patch from 1 pixel up the first axis away from the centre.
        If ``None``, then no offsets are applied.
    cval : `float`, optional
        The value outside the image boundaries.

    Returns
    -------
    patches : ``(n_center, n_offset, n_channels, patch_shape)`` `ndarray
        The extracted patches including any offsets if they were requested.

    Raises
    ------
    ValueError
        If pixels array is not 2D
    """
    if pixels.ndim != 3:
        raise ValueError('Only 2D images are supported but '
                         'found {}'.format(pixels.shape))

    n_offsets = offsets.shape[0] if offsets is not None else 1
    half_r, half_c = (patch_shape[0] / 2, patch_shape[1] / 2)
    # These are the offsets to the corners of the patches (hence the negative)
    corners = np.array([[-half_r, -half_c], [half_r, half_c]])
    if offsets is None:
        offsets = np.zeros([1, 2])

    patches = np.full([patch_centers.shape[0],
                       n_offsets,
                       pixels.shape[0],
                       patch_shape[0],
                       patch_shape[1]],
                      fill_value=cval,
                      dtype=pixels.dtype)
    # This is equivalent to nearest neighbour sampling per offset
    half_pixel = (np.array([patch_shape]) % 2) / 2
    patch_centers = patch_centers + half_pixel
    bounds = np.round(patch_centers[:, None, None, :] +
                      offsets[:, None, :] +
                      corners).astype(int)
    # Limit the points to exist inside the image boundaries
    pixel_bounds = np.clip(bounds, [0, 0], [pixels.shape[1:]])
    # Then compute the regions inside the patches that need to be filled
    patch_bounds = pixel_bounds - bounds

    # pixel_bounds will have shape (n_points, n_offsets, 2, 2)
    # Loop over the patches
    for i, (pix_off, patch_off) in enumerate(zip(pixel_bounds, patch_bounds)):
        # Loop over the offsets
        for j, (pix_b, patch_b) in enumerate(zip(pix_off, patch_off)):
            pix_slice = (slice(pix_b[0, 0], pix_b[1, 0]),
                         slice(pix_b[0, 1], pix_b[1, 1]))
            patch_slice = (slice(patch_b[0, 0], patch_shape[0] + patch_b[1, 0]),
                           slice(patch_b[0, 1], patch_shape[1] + patch_b[1, 1]))

            # Set only the pixels that were inside the image bounds
            patches[i, j, :, patch_slice[0], patch_slice[1]] = \
                pixels[:, pix_slice[0], pix_slice[1]]

    return patches


def set_patches(patches, pixels, patch_centers, offset, offset_index):
    r"""
    Set the values of a group of patches into the correct regions of a copy
    of this image. Given an array of patches and a set of patch centers,
    the patches' values are copied in the regions of the image that are
    centred on the coordinates of the given centers.

    The patches argument can have any of the two formats that are returned
    from the `extract_patches()` and `extract_patches_around_landmarks()`
    methods. Specifically it can be:

        1. ``(n_center, n_offset, self.n_channels, patch_shape)`` `ndarray`
        2. `list` of ``n_center * n_offset`` :map:`Image` objects

    Currently only 2D images are supported.

    Parameters
    ----------
    patches : `ndarray` or `list`
        The values of the patches.
        A ``(n_center, n_offset, self.n_channels, patch_shape)`` `ndarray`
    pixels : ``(n_channels, height, width)`` `ndarray``
        Pixel array to replace the patches within
    patch_centers : :map:`PointCloud`
        The centers to set the patches around.
    offset : `list` or `tuple` or ``(1, 2)`` `ndarray`
        The offset to apply on the patch centers within the image.
    offset_index : `int`
        The offset index within the provided `patches` argument, thus the
        index of the second dimension from which to sample.

    Raises
    ------
    ValueError
        If pixels array is not 2D
    """
    if pixels.ndim != 3:
        raise ValueError('Only 2D images are supported but '
                         'found {}'.format(pixels.shape))

    patch_shape = patches.shape[-2:]
    # the [L]ow offset is the floor of half the patch shape
    l_r, l_c = (int(patch_shape[0] // 2), int(patch_shape[1] // 2))
    # the [H]igh offset needs to be one pixel larger if the original
    # patch was odd
    h_r, h_c = (int(l_r + patch_shape[0] % 2), int(l_c + patch_shape[1] % 2))
    for patches_with_offsets, point in zip(patches, patch_centers):
        patch = patches_with_offsets[offset_index]
        p = point + offset[0]
        p_r = int(p[0])
        p_c = int(p[1])
        pixels[:, p_r - l_r:p_r + h_r, p_c - l_c:p_c + h_c] = patch



from .base import (
    Renderer, Viewable, LandmarkableViewable, viewwrapper, Menpo3dMissingError,
    MenpowidgetsMissingError, PointGraphViewer2d, LandmarkViewer2d,
    ImageViewer2d, ImageViewer, AlignmentViewer2d, GraphPlotter,
    view_image_landmarks, view_patches, plot_gaussian_ellipses, plot_curve)
from .textutils import (print_progress, progress_bar_str, print_dynamic,
                        bytes_str)
from .viewmatplotlib import MatplotlibRenderer



import numpy as np
from numpy.testing import assert_allclose
from pytest import raises
from mock import Mock

from menpo.transform import Transform


x = np.zeros([5, 5])


class MockTransform(Transform):
    def _apply(self, x, **kwargs):
        return np.array(x)


def test_transform_n_dims_return_None():
    tr = MockTransform()
    n_dims = tr.n_dims
    assert (n_dims is None)


def test_transform_n_dims_output_return_None():
    tr = MockTransform()
    n_dims = tr.n_dims_output
    assert (n_dims is None)


def test_transform_apply_x_not_transformable():
    tr = MockTransform()
    new_x = tr.apply(x)

    assert (new_x is not x)
    assert_allclose(new_x, x)


def test_transform_apply_x_transformable():
    mocked = Mock()
    mocked._transform.return_value = mocked
    tr = MockTransform()
    transformed_mock = tr.apply(mocked)

    assert (transformed_mock is mocked)
    assert mocked._transform.called


def test_transform_apply_inplace_x_transformable():
    mocked = Mock()
    mocked._transform_inplace.return_value = mocked
    tr = MockTransform()
    no_return = tr._apply_inplace(mocked)

    assert (no_return is None)
    assert mocked._transform_inplace.called


def test_transform_apply_inplace_x_not_transformable():
    tr = MockTransform()
    with raises(ValueError):
        tr._apply_inplace(x)


def test_transform_compose_before():
    mocked = Mock()
    mocked._apply.return_value = x
    tr = MockTransform()
    chain = tr.compose_before(mocked)
    # Check transform chain
    assert (len(chain.transforms) == 2)
    assert (chain.transforms[0] is tr)
    assert (chain.transforms[1] is mocked)


def test_transform_compose_after():
    mocked = Mock()
    mocked._apply.return_value = x
    tr = MockTransform()
    chain = tr.compose_after(mocked)
    # Check transform chain
    assert (len(chain.transforms) == 2)
    assert (chain.transforms[0] is mocked)
    assert (chain.transforms[1] is tr)



import sys

from setuptools import setup, find_packages

import versioneer

# Please see conda/meta.yaml for other binary dependencies
install_requires = ['numpy>=1.14',
                    'scipy>=1.0',
                    'matplotlib>=3.0',
                    'pillow>=4.0']

if sys.version_info.major == 2:
    install_requires.append('pathlib==1.0')

setup(
    name='menpo',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='A Python toolkit for handling annotated data',
    author='The Menpo Team',
    author_email='hello@menpo.org',
    packages=find_packages(),
    install_requires=install_requires,
    package_data={'menpo': ['data/*']},
    tests_require=['pytest>=5.0', 'mock>=3.0']
)



from menpo.base import MenpoMissingDependencyError


try:
    from .vlfeat import (dsift, fast_dsift, vector_128_dsift,
                         hellinger_vector_128_dsift)
except MenpoMissingDependencyError:
    pass

# Remove from scope
del MenpoMissingDependencyError
